#! /usr/bin/env python3


import re
import json
import logging
from urllib import request
from urllib.error import URLError

from . import constants
from . twitchobjs import (
    TwitchFollow,
    Game,
    UserData,
    Vod,
    Stream)


logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)
# logger.setLevel(logging.DEBUG)

file_handler = logging.FileHandler(constants.TWITCH_API_LOG)
formatter = logging.Formatter(constants.TWITCH_API_FORM)
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


class TwitchAPI():
    """Performs API calls to Twitch"""

    def __init__(self, oauth):
        self.__headers = {
            'Client-ID': 'm3ggy4wqrvs4sfhv996glushskv1fi',
            'Authorization': 'Bearer ' + oauth}
        self.__oauth = oauth

    def _makeUrl(self, names):
        """Returns values as request URL if values match regex"""

        first = self._twitch_dict['first']
        conf = self._twitch_dict['cont']
        names_ext = first + conf.join(names)
        in_url = self._twitch_dict['url'] + names_ext

        reg = self._twitch_dict['reg']
        r = re.compile(reg)
        match = r.match(in_url)

        if match:
            url = match.group()
            return url

    def request(self, url):
        """Perform Twitch request call"""

        # redundancy regex for URL
        r = re.compile(r'(https?://)([a-z]{1,3}\.)?(twitch.tv/)(.*)')
        reg_match = r.match(url)
        if not reg_match:
            logger.warning('[REQUEST] Twitch request received unexpected URL')
            raise ValueError('Twitch request received unexpected URL')

        try:
            h = request.Request(url, headers=self.__headers)
            u = request.urlopen(h)
        except URLError as e:
            logger.warning(f'Error using TwitchAPI: {e}')
            print("Error using Twitch API: ", e)
            return None

        j = json.loads(u.read().decode('utf-8'))
        logger.debug(f'[REQUEST] Headers: {self.__headers}')
        logger.debug(f'[REQUEST] URL: {url}')
        logger.debug(f'[REQUEST] Answer: {j}')
        return j


class TwitchValidate(TwitchAPI):
    """Give validation values for TwitchAPI()"""

    __url = 'https://id.twitch.tv/oauth2/validate'

    def __init__(self, oauth):
        logger.debug('[VALIDATE] get key validation')
        super().__init__(oauth)
        self.__data = self.request(self.__url)

    def get(self):
        """Return validation answer"""

        return self.__data


class TwitchLoadFollow(TwitchAPI):
    """Give check following values for TwitchAPI()"""

    _twitch_dict = {
        'url': 'https://api.twitch.tv/helix/users/follows',
        'first': '?from_id=',
        'cont': '',
        'reg': (
            r'(https://api.twitch.tv/helix/users/follows)'
            r'(\?from_id=[0-9]*$)')}

    def __init__(self, oauth, id_):
        logger.debug('[FOLLOWS] load follows from user')
        super().__init__(oauth)
        url = self._makeUrl(id_)
        self.__data = self.request(url)

    def get(self):
        """Return user following answer"""

        twitch_follows = tuple(
            TwitchFollow(raw=r) for r in self.__data['data'])
        return twitch_follows


class TwitchGameData(TwitchAPI):
    """Give game data values for TwitchAPI()"""

    _twitch_dict = {
        'url': 'https://api.twitch.tv/helix/games',
        'first': '?id=',
        'cont': '&id=',
        'reg': (
            r'^(https://api.twitch.tv/helix/games)'
            r'(\?id=[0-9]+)'
            r'(&id=[0-9]+)*$')}

    def __init__(self, oauth, ids):
        logger.debug('[GAME] get game data')
        super().__init__(oauth)
        url = self._makeUrl(ids)
        self.__data = self.request(url)

    def get(self):
        """Return game data answer"""

        games = tuple(Game(raw=r) for r in self.__data['data'])
        return games


class TwitchVod(TwitchAPI):
    """Give vod values for TwitchAPI()"""

    _twitch_dict = {
        'url': 'https://api.twitch.tv/helix/videos',
        'first': '?user_id=',
        'cont': '',
        'reg': (
            r'^(https://api.twitch.tv/helix/videos)'
            r'(\?user_id=[0-9]+)$')}

    def __init__(self, oauth, ids):
        logger.debug('[VOD] get vods')
        super().__init__(oauth)
        url = self._makeUrl(ids)
        self.__data = self.request(url) or {}

    def get(self):
        """Return vods for users answer"""

        vods = tuple(Vod(raw=r) for r in self.__data.get('data') or {})
        return vods


class TwitchStream(TwitchAPI):
    """Give stream values for TwitchAPI()"""

    _twitch_dict = {
        'url': 'https://api.twitch.tv/helix/streams',
        'first': '?user_login=',
        'cont': '&user_login=',
        'reg': (
            r'^(https://api.twitch.tv/helix/streams)'
            r'(\?user_login=[a-z0-9_]+)'
            r'(&user_login=[a-z0-9_]+)*$')}

    def __init__(self, oauth, names):
        logger.debug('[STREAM] get streaming')
        super().__init__(oauth)
        url = self._makeUrl(names)
        self.__data = self.request(url) or {}

    def get(self):
        """Return users streaming answer"""

        streams = tuple(Stream(raw=r) for r in self.__data.get('data') or {})
        return streams


class TwitchUserData(TwitchAPI):
    """Give user data values for TwitchAPI()"""

    _twitch_dict = {
        'url': 'https://api.twitch.tv/helix/users',
        'first': '?login=',
        'cont': '&login=',
        'reg': (
            r'^(https://api.twitch.tv/helix/users)'
            r'(\?login=[a-z0-9_]+)'
            r'(&login=[a-z0-9_]+)*$')}

    def __init__(self, oauth, names):
        logger.debug('[USER] get user data')
        super().__init__(oauth)
        url = self._makeUrl(names)
        self.__data = self.request(url)

    def get(self):
        """Return user data answer"""

        users = list(UserData(raw=r) for r in self.__data['data'])
        return users



#! /usr/bin/env python3


import os
import logging
from pathlib import Path


__version__ = '0.20.0'
__author__ = 'obtusescholar'

BASES = {
    'posix': '~/.config/StreamerRetriever/',
    'nt': '~/Documents/StreamerRetriever/'}


def getBase():
    str_path = BASES[os.name]
    abs_path = Path.resolve(Path.expanduser(Path(str_path)))
    return abs_path


BASE = getBase()

FOLLOWS = 'follows.csv'
DB = 'streamerretriever.json'
# TWITCH = 'twitch.csv'
TWITCH = 'twitch.json'
GAME_CACHE = 'game_cache.json'
CONFIG = 'cli_conf.json'
# STREAM_LOG = 'stream.log'
# CLI_LOG = 'cli.log'
# FO_LOG = 'fo.log'
# TOOLS_LOG = 'tools.log'


def getLog(log, key):
    logs = {
        'unified': {
            # 'file': 'streamerretriever_cli.log'},
            'file': 'streamerretriever.log'},
        'stream': {
            'file': 'cli_stream.log',
            'tag':  '[STREAM]'},
        'cli': {
            'file': 'cli_cli.log',
            'tag':  '[CLI]'},
        'fo': {
            'file': 'cli_fo.log',
            'tag':  '[FO]'},
        'tools': {
            'file': 'cli_tools.log',
            'tag':  '[TOOLS]'},
        'twitch_request': {
            'file': 'twitch_request.log',
            'tag': '[TWITCH_REQUEST]'},
        'twitch_obj': {
            'file': 'twitch_obj.log',
            'tab': '[TWITCH_OBJ]'}}
    return logs[log][key]


def getLogFile(log):
    # folder = 'logs/'

    # log_folder = Path(BASES[os.name]) / Path(folder)
    log_folder = Path(BASES[os.name])
    # rel_path = log_folder / Path(getLog(log, 'file'))
    rel_path = log_folder / Path(getLog('unified', 'file'))
    abs_path = Path.resolve(Path.expanduser(rel_path))

    return abs_path


LOG_LEVEL = logging.WARNING
STREAM_LOG = getLogFile('tools')
CLI_LOG = getLogFile('gui')
FO_LOG = getLogFile('gui_tools')
TOOLS_LOG = getLogFile('follows')
TWITCH_API_LOG = getLogFile('twitch_request')
TWITCH_OBJ_LOG = getLogFile('twitch_request')


def getLogForm(log):
    log_form = '[%(asctime)s] [%(levelname)s] %(message)s'
    logger_form = '[CLI]' + ' ' + getLog(log, 'tag') + ' ' + log_form
    return logger_form


STREAM_FORM = getLogForm('stream')
CLI_FORM = getLogForm('cli')
FO_FORM = getLogForm('fo')
TOOLS_FORM = getLogForm('tools')
TWITCH_API_FORM = getLogForm('twitch_request')
TWITCH_OBJ_FORM = getLogForm('twitch_request')



#! /usr/bin/env python3


import os
import re
import csv
import json
import logging
from pathlib import Path

from . import constants


logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

file_handler = logging.FileHandler(constants.FO_LOG)
formatter = logging.Formatter(constants.FO_FORM)
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


class FileOperator():
    """Source file that contains items."""

    def __init__(self, path, filetype=False):
        """Initializes path."""

        logger.debug(f'[INIT] Path is: "{path}"')
        self.__path = self.__resolvePath(path)
        self.__checkFile(self.__path)
        self.__filetype = self.__getFiletype(filetype)

    def __getFiletype(self, filetype):
        """Tries to guess the filetype from filename."""

        if filetype:
            ext = filetype
            logger.debug('[GET] File extension provided')
        else:
            reg = re.compile(r'.*\.([^\s^\.]+)$')
            if reg.match(str(self.__path)):
                ext = reg.match(str(self.__path)).groups()[0]
                logger.debug('[GET] Guessed file extension')
            else:
                ext = False
                logger.debug('[GET] File extension not resolved')
        logger.debug(f'[GET] File extension: "{ext}"')
        return ext

    def __resolvePath(self, path):
        """Resolves path OS independently."""

        p = Path.resolve(Path.expanduser(Path(path)))
        logger.debug(f'[RESOLVE] Given absolute path: "{p}"')
        return p

    def __checkFile(self, path):
        """Checks that file exists and creates it and path."""

        if not Path.exists(path):
            base = path.parents[0]
            if not Path.exists(base):
                os.makedirs(base)
                logger.info(f'[CHECK] Created directory "{base}" for sources.')

            with open(path, 'a'):
                pass
            logger.info(f'[CHECK] Created file "{path}" for sources.')
        else:
            logger.debug(f'[CHECK] "{path}" found.')

    def __fileEmpty(self, path):
        """Checks if file is empty."""

        if os.stat(path).st_size == 0:
            return True
        return False

    def loadFile(self, singleLayer=False):
        """Loads data from file.

        If given or guessed filetype is found then correct method is
        used to load data from file.

        Emptry string is returned if file is empty.

        Known filetypes: json, txt, csv.
        """

        if self.__fileEmpty(self.__path):
            logger.debug('[LOAD] File is empty')
            return ''

        if self.__filetype == 'json':
            with open(self.__path, 'r') as f:
                data = json.load(f)
            logger.debug('[LOAD] json loaded')

        elif self.__filetype == 'txt':
            with open(self.__path, 'r') as f:
                data = f.read()
            logger.debug('[LOAD] txt loaded')

        elif self.__filetype == 'csv':
            with open(self.__path, 'r', newline='') as f:
                csv_data = csv.reader(f, delimiter=',')
                data = list(csv_data)
            if singleLayer:
                data = [item for sublist in data for item in sublist]
            logger.debug(f'[LOAD] csv loaded, singleLayer={singleLayer}')

        # only for extreme debug situations
        # logger.debug(f'[LOAD] data: "{data}"')
        return data

    def saveFile(self, data=False):
        """Saves data to file.

        If given or guessed filetype is found then correct method is
        used to write given data.

        Empty data can be written to any filetype.

        Known filetypes: json, txt, csv.
        """

        if data:
            if self.__filetype == 'json':
                with open(self.__path, 'w') as f:
                    json.dump(data, f)

            elif self.__filetype == 'txt':
                with open(self.__path, 'w') as f:
                    f.write(data)

            elif self.__filetype == 'csv':
                with open(self.__path, 'w', newline='') as f:
                    writer = csv.writer(f)
                    writer.writerow(data)
            logger.debug(f'[SAVE] {self.__filetype} saved')

        else:
            with open(self.__path, 'w') as f:
                f.write('')
            logger.debug(f'[SAVE] "" to: "{self.__path}"')



#! /usr/bin/env python3

import logging
import datetime

from . import constants


logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

file_handler = logging.FileHandler(constants.TWITCH_OBJ_LOG)
formatter = logging.Formatter(constants.TWITCH_OBJ_FORM)
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


class EmptyValue():
    """Define empty values if Twitch returns 'Null'"""

    def _noGameId(self):
        """If game_id is not found return ID: 0"""

        empty = {'game_id': 0}
        self._data.update(empty)


class Merge():
    """Merges two objects and their data"""

    @staticmethod
    def _merge(merge_items, item, key):
        for merge_item in merge_items:
            if merge_item == item:
                ob = dict(item.__dict__[key])
                ob.update(dict(merge_item.__dict__[key]))
        return ob


class User():
    _keys = {
        'user_id': {
            'twitch': 'id',
            'pretty': 'ID'},
        'display_name': {
            'twitch': 'display_name',
            'pretty': 'Display Name'}}
    _key_source = 'user'

    def __init__(self, *ignore, raw):
        self._keys = dict(self._keys)
        if raw:
            self._data = self.__genData(raw)

    def __hash__(self):
        h = int(self._data['user_id'])
        return h

    def __eq__(self, other):
        if isinstance(other, User):
            return hash(other) == self.__hash__()
        else:
            return NotImplemented

    def __repr__(self):
        return f'{self.__class__.__name__}({self._data})'

    def __genData(self, raw):
        data = {}
        for k in self._keys:
            data[k] = raw[self._keys[k]['twitch']]
        return data

    def get(self, key):
        return self._data[key]

    def getKeys(self):
        # keys = tuple(self._keys)
        return self._keys

    def getUserId(self):
        return self.get('user_id')

    def getDispName(self):
        return self.get('display_name')

    def getKeySource(self):
        return self._key_source

    # def getPretty(self, keys):
    #     pretty = []
    #     for key in keys:
    #         if key in self._keys:
    #             pretty.append(self._keys[key]['pretty'])
    #     return pretty


class TwitchFollow(User):
    __follow_keys = {
        'user_id': {
            'twitch': 'to_id'},
        'name': {
            'twitch': 'to_login'},
        'display_name': {
            'twitch': 'to_name'},
        'followed_from': {
            'twitch': 'followed_at'}}
    _key_source = 'twitch_follow'

    def __init__(self, *ignore, raw=False, data=False):
        self._keys = self.__follow_keys
        if raw and not data:
            super().__init__(raw=raw)
        elif data and not raw:
            self._data = data

    def __hash__(self):
        h = int(self._data['user_id'])
        return h

    def __eq__(self, other):
        if isinstance(other, User):
            return other.getId() == self.__hash__()
        else:
            return NotImplemented

    def getId(self):
        return self.get('user_id')

    def getName(self):
        return self.get('name')

    def getFollowedFrom(self):
        return self.get('followed_from')


class Game(User):
    __game_keys = {
        'box_art_url': {
            'twitch': 'box_art_url',
            'pretty': 'Box Art URL'},
        'game_id': {
            'twitch': 'id',
            'pretty': 'Game ID'},
        'game': {
            'twitch': 'name',
            'pretty': 'Game'}}
    _key_source = 'game'

    def __init__(self, *ignore, raw=False, data=False):
        self._keys = self.__game_keys
        if raw and not data:
            super().__init__(raw=raw)
            if not self.getGameId():
                self._noGameId()
        elif data and not raw:
            self._data = data

        # else:
        #     self._data = {
        #         'box_art_url': 'Unknown',
        #         'game_id': '0',
        #         'game': 'Unknown'}

    def __hash__(self):
        h = int(self._data['game_id'])
        return h

    def __eq__(self, other):
        if isinstance(other, User):
            return other.getGameId() == self.__hash__()
        else:
            return NotImplemented

    def getBoxArtUrl(self):
        return self.get('box_art_url')

    def getGameId(self):
        return self.get('game_id')

    def getGame(self):
        return self.get('game')


class UserData(User):
    __offline_keys = {
        'name': {
            'twitch': 'login',
            'pretty': 'Name'},
        'type': {
            'twitch': 'type',
            'pretty': 'Type'},
        'partner': {
            'twitch': 'broadcaster_type',
            'pretty': 'Twitch Status'},
        'description': {
            'twitch': 'description',
            'pretty': 'Description'},
        'icon_url': {
            'twitch': 'profile_image_url',
            'pretty': 'Icon URL'},
        'offline_image': {
            'twitch': 'offline_image_url',
            'pretty': 'Offline Image'},
        'view_count': {
            'twitch': 'view_count',
            'pretty': 'View Count'}}
    _key_source = 'userdata'

    def __init__(self, *ignore, raw=False, data=False):
        self._keys = dict(self._keys)
        if raw and not data:
            # self._keys = dict(self._keys())
            self._keys.update(self.__offline_keys)
            super().__init__(raw=raw)
        elif data and not raw:
            self._data = data
        else:
            # self._keys = dict(self._keys())
            self._keys.update(self.__offline_keys)
            # m = 'UserData __init__() requires either "raw" or "data" keyword'
            # raise TypeError(m)

    def getName(self):
        return self.get('name')

    def getType(self):
        return self.get('type')

    def getPartner(self):
        return self.get('partner')

    def getDesc(self):
        return self.get('description')

    def getIconUrl(self):
        return self.get('icon_url')

    def getOfflineImage(self):
        return self.get('offline_image')

    def getViewCount(self):
        return self.get('view_count')


class Vod(User):
    __vod_keys = {
        'vod_id': {
            'twitch': 'id',
            'pretty': 'Vod ID'},
        'user_id': {
            'twitch': 'user_id',
            'pretty': 'User ID'},
        'display_name': {
            'twitch': 'user_name',
            'pretty': 'Display Name'},
        'vod_title': {
            'twitch': 'title',
            'pretty': 'Vod Title'},
        'vod_description': {
            'twitch': 'description',
            'pretty': 'Vod Description'},
        'created_at': {
            'twitch': 'created_at',
            'pretty': 'Created At'},
        'published_at': {
            'twitch': 'published_at',
            'pretty': 'Published At'},
        'vod_url': {
            'twitch': 'url',
            'pretty': 'Vod URL'},
        'thumbnail_url': {
            'twitch': 'thumbnail_url',
            'pretty': 'Thumbnail URL'},
        'viewable': {
            'twitch': 'viewable',
            'pretty': 'Viewable'},
        'vod_view_count': {
            'twitch': 'view_count',
            'pretty': 'Vod View Count'},
        'language': {
            'twitch': 'language',
            'pretty': 'Language'},
        'vod_type': {
            'twitch': 'type',
            'pretty': 'Vod Type'},
        'duration': {
            'twitch': 'duration',
            'pretty': 'Duration'}}
    _key_source = 'vod'

    def __init__(self, *ignore, raw):
        self._keys = dict(self._keys)
        self._keys.update(self.__vod_keys)
        if raw:
            super().__init__(raw=raw)

    def getVodId(self):
        return self.get('vod_id')

    def getTitle(self):
        return self.get('vod_title')

    def getVideoDesc(self):
        return self.get('vod_description')

    def getCreatedAt(self):
        return self.get('created_at')

    def getPublishedAt(self):
        return self.get('published_at')

    def getVodUrl(self):
        return self.get('vod_url')

    def getThumbUrl(self):
        return self.get('thumbnail_url')

    def getViewable(self):
        return self.get('viewable')

    def getViewCount(self):
        return self.get('vod_view_count')

    def getLanguage(self):
        return self.get('language')

    def getType(self):
        return self.get('type')

    def getDuration(self):
        return self.get('duration')


class VodFullData(UserData, Vod):
    pass


class VodUser(VodFullData, Vod, Merge):
    def __init__(self, db, stream):
        self._keys = dict(self._keys)
        if db and stream:
            self._data = self._merge(db, stream, '_data')
        else:
            super().__init__(data=[])


class Stream(User, EmptyValue):
    __streaming_keys = {
        'user_id': {
            'twitch': 'user_id',
            'pretty': 'User ID'},
        'display_name': {
            'twitch': 'user_name',
            'pretty': 'Display Name'},
        'stream_id': {
            'twitch': 'id',
            'pretty': 'Stream ID'},
        'game_id': {
            'twitch': 'game_id',
            'pretty': 'Game ID'},
        'title': {
            'twitch': 'title',
            'pretty': 'Title'},
        'viewers': {
            'twitch': 'viewer_count',
            'pretty': 'Viewers'},
        'started_at': {
            'twitch': 'started_at',
            'pretty': 'Started At'},
        'thumbnail_url': {
            'twitch': 'thumbnail_url',
            'pretty': 'Thumbnail URL'},
        'tag_ids': {
            'twitch': 'tag_ids',
            'pretty': 'Tag IDs'}}
    _key_source = 'stream'

    def __init__(self, *ignore, raw):
        # self._keys = dict(self._keys)
        self._keys = dict(self._keys)
        self._keys.update(self.__streaming_keys)
        if raw:
            super().__init__(raw=raw)
            self.__insertTime()
            if not self.getGameId():
                self._noGameId()
        else:
            self._keys.update({'time': {'twitch': 'time', 'pretty': 'Time'}})

        # if not self.getGameId():
        #     empty = {'game_id': 0}
        #     self._data.update(empty)

    def __insertTime(self):
        started = self.getStarted()
        time_format = '%Y-%m-%dT%H:%M:%SZ'
        started_stamp = datetime.datetime.strptime(started, time_format)
        now = datetime.datetime.utcnow()
        delta = now - started_stamp
        time = str(datetime.timedelta(seconds=delta.seconds))
        self._data['time'] = time

    def getStreamId(self):
        return self.get('stream_id')

    def getGameId(self):
        return self.get('game_id')

    def getTitle(self):
        return self.get('title')

    def getViewers(self):
        return self.get('viewers')

    def getStarted(self):
        return self.get('started_at')

    def getThumbUrl(self):
        return self.get('thumbnail_url')

    def getTagIds(self):
        return self.get('tag_ids')

    def getTime(self):
        return self.get('time')


class CombineUserStream(UserData, Stream):
    pass


class StreamingUser(CombineUserStream, Stream, Merge):
    def __init__(self, db, stream):
        self._keys = dict(self._keys)
        if db and stream:
            self._data = self._merge(db, stream, '_data')
        else:
            super().__init__(data=[])

    # def get(self, key):
    #     def tryValue(key, i):
    #         try:
    #             value = self.__sources[i].get(key)
    #         except KeyError:
    #             return tryValue(key, i+1)
    #         return value
    #     return tryValue(key, 0)


class StreamingUserGame(StreamingUser, Game):
    def __init__(self, games, stream_user):
        self._keys = dict(self._keys)
        if games and stream_user:
            self._data = self._merge(games, stream_user, '_data')
        else:
            super().__init__(data=[])

    def __hash__(self):
        h = int(self._data['user_id'])
        return h

    def __eq__(self, other):
        if isinstance(other, StreamingUserGame):
            return hash(other) == self.__hash__()
        else:
            return NotImplemented

    @staticmethod
    def _merge(merge_items, item, key):
        for merge_item in merge_items:
            if hash(merge_item) == int(item.getGameId()):
                ob = dict(item.__dict__[key])
                ob.update(dict(merge_item.__dict__[key]))
        return ob



#! /usr/bin/env python3


import os
import re
import logging

from . import constants
from . twitchrequest import (
    TwitchValidate,
    TwitchLoadFollow,
    TwitchGameData,
    TwitchVod,
    TwitchStream,
    TwitchUserData)
from . twitchobjs import (
    Game,
    UserData,
    Vod,
    VodUser,
    Stream,
    StreamingUser,
    StreamingUserGame)
from . fileoperator import FileOperator


logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

file_handler = logging.FileHandler(constants.STREAM_LOG)
formatter = logging.Formatter(constants.STREAM_FORM)
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


class Key():
    def __init__(self, index, key):
        self.__data = key
        self.__name = index

    def __eq__(self, other):
        if isinstance(other, Key):
            return other.getName() == self.getName()
        elif isinstance(other, str):
            return other == self.getName()
        else:
            return NotImplemented

    def __repr__(self):
        return f'{self.__class__.__name__}({{{self.__name}: {self.__data}}})'

    def getName(self):
        return self.__name

    def getDispName(self):
        return self.__data['pretty']


class StreamerRetriever():
    __base = constants.BASES[os.name]

    def __init__(self, info):
        # self.__oauth = info.getToken()
        self.__info = info
        self.__loadFiles()
        self.__upDb()

        u = UserData()
        s = Stream(raw=False)
        g = Game(raw=False)
        v = Vod(raw=False)
        sources = [u, s, g, v]
        self.__keys = {}
        for source in sources:
            self.__keys[source.getKeySource()] = dict(source.getKeys())

    def __loadFiles(self):
        b = self.__base
        self.__follows_fo = FileOperator(b + constants.FOLLOWS)
        self.__follows = self.__follows_fo.loadFile(singleLayer=True)

        self.__game_cache_fo = FileOperator(b + constants.GAME_CACHE)
        game_cache = self.__game_cache_fo.loadFile() or []
        self.__game_cache = list(Game(data=g) for g in game_cache)
        self.__cacheEmptyGame()

        # games = tuple(Game(raw=r) for r in self.__data['data'])

        self.__db_fo = FileOperator(b + constants.DB)
        dbs = self.__db_fo.loadFile()
        self.__users = tuple(UserData(data=db) for db in dbs)

    def __cacheGames(self, items):
        game_ids = set((i.getGameId() for i in items))
        cache_ids = set((c.getGameId() for c in self.__game_cache))
        new_ids = game_ids - cache_ids

        # apparently some games come as empty values
        # update = set(new if new else '0' for new in new_ids)
        update = set(new for new in new_ids if new)

        if update and update is not set():
            twitch_games = TwitchGameData(self.__info.getToken(), update)
            update_games = twitch_games.get()
            self.__game_cache.extend(update_games)
            self.__saveGameCache()

    def __cacheEmptyGame(self):
        for game in self.__game_cache:
            if hash(game) == 0:
                break
        else:
            empty = {
                'box_art_url': 'Unknown',
                'game_id': '0',
                'game': 'Unknown'}
            empty_game = Game(data=empty)
            self.__game_cache.append(empty_game)
            self.__saveGameCache()

    def __stream(self):
        if self.__follows:
            twitch_streaming = TwitchStream(
                self.__info.getToken(), self.__follows)
            streaming = twitch_streaming.get()
        else:
            streaming = []
        self.__cacheGames(streaming)
        return streaming

    def __getStreaming(self):
        stream = self.__stream()
        streaming_user = tuple(StreamingUser(self.__users, x) for x in stream)
        streaming_data = tuple(
            StreamingUserGame(self.__game_cache, y) for y in streaming_user)
        return streaming_data

    def __updateUsers(self):
        twitch_user_data = TwitchUserData(
            self.__info.getToken(), self.__follows)
        user_data = twitch_user_data.get()

        sorted_data = []
        for follow in self.__follows:
            for user in user_data:
                if user.getName() == follow:
                    sorted_data.append(user)

        self.__users = sorted_data

    def getFiles(self):
        files = {
            'follows': self.__follows_fo,
            'db': self.__db_fo,
            'game_cache': self.__game_cache_fo}
        return files

    def __saveGameCache(self):
        u = []
        for game in self.__game_cache:
            u.append(game.__dict__['_data'])
        self.__game_cache_fo.saveFile(u)

    def __saveDb(self):
        u = []
        for user in self.__users:
            u.append(user.__dict__['_data'])
        self.__db_fo.saveFile(u)

    # def __addRemoveDb(self):
    #     logger.info(
    #         f'Update DB removed: {set(db_follows) - set(self.__follows)}')
    #     logger.info(
    #         f'Update DB added: {set(self.__follows) - set(db_follows)}')
    #     new = set(self.__follows)
    #     old = set(db_follows)
    #     if new - old == set():
    #         self.__dbSorting()
    #     # elif new - old != new:
    #     #     self.__dbSorting()
    #     #     self.__updateUsers()
    #     else:
    #         self.__updateUsers()
    #     self.__saveDb()

    def __upDb(self):
        db_follows = []
        for f in self.__users:
            name = f.getName()
            db_follows.append(name)

        if set(db_follows) != set(self.__follows):
            logger.info(
                f'Update DB removed: {set(db_follows) - set(self.__follows)}')
            logger.info(
                f'Update DB added: {set(self.__follows) - set(db_follows)}')
            new = set(self.__follows)
            old = set(db_follows)
            if new - old == set():
                self.__dbSorting()
            # elif new - old != new:
            #     self.__dbSorting()
            #     self.__updateUsers()
            else:
                self.__updateUsers()
            self.__saveDb()
        elif db_follows != self.__follows:
            self.__dbSorting()

    def __dbSorting(self):
        db_sorted = []
        for f in self.__follows:
            for u in self.__users:
                if u.getName() == f:
                    db_sorted.append(u)
                    break
        self.__users = db_sorted
        self.__saveDb()
        logger.info(
            f'Update DB sorting: {tuple(e.getName() for e in db_sorted)}')

    def __saveFollows(self):
        self.__follows_fo.saveFile(self.__follows)

    def __cacheGameId(self, game_id):
        if game_id in self.__game_cache:
            pass

    def updateValidate(self):
        twitchValidate = TwitchValidate(self.__info.getToken())
        validation = twitchValidate.get()
        self.__info.setValidate(validation)

    def getValidate(self):
        return self.__info

    def getTwitchFollows(self):
        id_ = self.__info.getId()
        if not id_:
            self.updateValidate()
            id_ = self.__info.getId()
        twitch_load_follow = TwitchLoadFollow(self.__info.getToken(), id_)
        twitch_follows = twitch_load_follow.get()
        # follows = tuple(twitch.getName() for twitch in twitch_follows)
        # return follows
        return twitch_follows

    def updateDb(self, follows):
        self.__follows = follows
        self.__saveFollows()
        self.__upDb()

    def getDb(self):
        return self.__users

    def getPretty(self, keys):
        pretty = []
        for key in keys:
            for source in self.__keys:
                if key in self.__keys[source]:
                    pretty.append(self.__keys[key]['pretty'])
                else:
                    pretty.append(key)
        return pretty

    def getKeys(self, source=None):
        keys = {}
        if not source:
            for s in self.__keys:
                keys.update(self.__keys[s])
            return keys
        else:
            for s in source:
                keys.update(self.__keys[s])
            return keys

    def getKeyObjs(self, source=None):
        keys = self.getKeys(source)
        key_objs = list(Key(key, keys[key]) for key in keys)
        return key_objs

    def getOnline(self):
        return self.__getStreaming()

    def getVod(self, user_id):
        twitch_vods = TwitchVod(self.__info.getToken(), user_id)
        vods = twitch_vods.get()
        vod_user = tuple(VodUser(self.__users, x) for x in vods)
        return vod_user


if __name__ == '__main__':
    def oauth():
        import webbrowser

        base = constants.BASES[os.name]
        f = constants.TWITCH
        p = base + f

        fo = FileOperator(p)

        def __generateOAuth():
            url = (
                'https://id.twitch.tv/oauth2/authorize'
                '?client_id=m3ggy4wqrvs4sfhv996glushskv1fi'
                '&redirect_uri=http://localhost'
                '&response_type=token')
            webbrowser.open(url)

        def getKey():
            oauth = fo.loadFile(singleLayer=True)
            if oauth:
                return oauth[0]
            else:
                __generateOAuth()
                print('Paste resulting URL after doing Twitch login')
                url = input()

                r = re.compile(
                    r'^(http://localhost/)'
                    r'(\#access_token=([a-z0-9]+))'
                    r'(&scope=&token_type=bearer)$')
                match = r.match(url)

                if match:
                    oauth = match.groups()[2]
                    return oauth

        return getKey()

    def testingStreamerRetriever():
        pass
        # print(s.getDb())

        # print()
        # print(s.getOnline())

        # print()
        # print(StreamingUser({}, {}).getKeys())

        # print()
        # print(Stream(raw='').getKeys())

        # print()
        # users = []
        # for i, user in enumerate(s.getDb()):
        #     users.append(user)
        #     print(f'#{i+1} ' + user.getDispName())

        # user_in = int(input('#')) - 1
        # user_id = users[user_in].getUserId()
        # print(s.getVod(user_id))

    def testing():
        token = oauth()
        names = ['esl_csgo', 'esamarathon', 'gamesdonequick']

        twitch_user_data = TwitchUserData(token, names)
        user_data = twitch_user_data.get()
        for user in user_data:
            print(hash(user))
            print(user)
            print()

        twitch_streaming = TwitchStream(token, names)
        streaming = twitch_streaming.get()
        for stream in streaming:
            print(hash(stream))
            print(stream)
            print()

        # stream_users = tuple(StreamingUser(user_data, x) for x in streaming)
        # for s in stream_users:
        #     print(hash(s))
        #     print(s)
        #     print()

    def testUpdateDb():
        token = oauth()
        s = StreamerRetriever(token)
        follows = ['esamarathon', 'esl_csgo']
        s.updateDb(follows)

    token = oauth()
    # s = StreamerRetriever(token)
    # online = s.getOnline()
    # for o in online:
    #     # print(o)
    #     # print(o.__dict__)
    #     # print(o.getName() + ' - ' + o.getGame())
    #     pass

    def validate():
        twitch_validate = TwitchValidate(token)
        print(twitch_validate)

    # testing()
    # testingStreamerRetriever()
    # testUpdateDb()



#! /usr/bin/env python3


import os
import re
import logging
import argparse
import datetime
import webbrowser
import subprocess
from pathlib import Path

from . import tools
from . import constants
from . fileoperator import FileOperator
from . streamerretriever import StreamerRetriever


logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

file_handler = logging.FileHandler(constants.CLI_LOG)
formatter = logging.Formatter(constants.CLI_FORM)
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


def argParser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v', '--version', help='show version', action='store_true')
    parser.add_argument(
        '-c', '--check', help='check online streams and output names',
        action='store_true')
    parser.add_argument(
        '-l', '--link', help='check online streams and output links',
        action='store_true')
    parser.add_argument(
        '-p', '--play', help='check streams and choose what to play',
        action='store_true')
    parser.add_argument(
        '-d', '--vod', help='check vods and choose what to play',
        action='store_true')
    parser.add_argument(
        '-w', '--web', help='use default webbrowser with play',
        action='store_true')
    parser.add_argument(
        '-m', '--mobile', help='make link into twitch mobile version',
        action='store_true')
    parser.add_argument(
        '-t', '--twitch-follows', help='print your follows from twitch',
        action='store_true')
    args = parser.parse_args()
    return args


def tokenInput():
    url = input()
    return url


def initStreamerRetriever():
    token = tools.oauth(tokenInput)
    streamerretriever = StreamerRetriever(token)
    return streamerretriever


def showVodInfo(value):
    def timeFormat():
        time_format = '%Y-%m-%dT%H:%M:%SZ'
        time = datetime.datetime.strptime(value.getCreatedAt(), time_format)
        return time.strftime("%a, %d. %B %H:%M")

    title = value.getTitle()
    show_info = (
        f'{timeFormat()}, Duration: {value.getDuration()}\n' + ' ' * 5 +
        f'{title}' + '\n')
    return show_info


def showDispName(value):
    return value.getDispName()


def showAsk(listing, showInfo):
    for i, value in enumerate(listing):
        t = '{:<5}'.format(f'#{i+1}')
        print(f'{t}{showInfo(value)}')


def checkAsk():
    val = input('#')
    r = re.compile(r'^([1-9][0-9]*)$')
    match = r.match(val)
    return match


def getAsk(listing, match):
    choice = int(match.group())
    if choice <= len(listing):
        value = listing[choice-1]
        return value


def askInput(listing, showInfo):
    showAsk(listing, showInfo)
    match = checkAsk()
    if match:
        value = getAsk(listing, match)
        return value


def vodPrinting(name):
    space = ' ' * 5
    prefix = '#   '
    txt = ' - VODS'
    line = '#' * (len(name.getDispName()) + len(txt) + len(prefix) * 2)
    print()
    print(space + line)
    print(space + prefix + name.getDispName() + txt + prefix[::-1])
    print(space + line)


def getPlayer():
    base = constants.BASES[os.name]
    cli_conf = constants.CONFIG
    p = Path.resolve(Path.expanduser(Path(base)))
    path = p / cli_conf

    fo = FileOperator(str(path))
    configs = fo.loadFile()
    player = configs.get('player')

    return player


def arguments():
    args = argParser()

    normal = 'https://www.twitch.tv/'
    mobile = 'https://m.twitch.tv/'

    vod_normal = normal + 'videos/'
    vod_mobile = mobile + 'videos/'

    def checkLinkMobile():
        streamerretriever = initStreamerRetriever()
        online = streamerretriever.getOnline()
        if args.check:
            for o in online:
                print(o.getDispName())
        elif args.mobile:
            for o in online:
                print(mobile + o.getName())
        else:
            for o in online:
                print(normal + o.getName())

    def playWeb():
        streamerretriever = initStreamerRetriever()

        online = streamerretriever.getOnline()
        if not online:
            return
        value = askInput(online, showDispName)
        if not value:
            return

        if args.mobile and args.web:
            url = mobile + value.getName()
            logger.info(f'Twitch URL: {url}')
            webbrowser.open(url)
        elif args.web:
            url = normal + value.getName()
            logger.info(f'Mobile URL: {url}')
            webbrowser.open(url)
        elif args.mobile:
            url = mobile + value.getName()
            command = getPlayer()
            command.append(url)
            logger.info(f'Mobile Command: {command}')
            subprocess.Popen(command, close_fds=True)
        else:
            url = normal + value.getName()
            command = getPlayer()
            command.append(url)
            logger.info(f'Twitch Command: {command}')
            subprocess.Popen(command, close_fds=True)

    def playVod():
        streamerretriever = initStreamerRetriever()

        database = streamerretriever.getDb()
        if not database:
            return
        value = askInput(database, showDispName)
        if not value:
            return

        vodPrinting(value)

        vods = streamerretriever.getVod(value.getUserId())
        if not vods:
            return
        chosen_vod = askInput(vods, showVodInfo)
        if not chosen_vod:
            return

        if args.mobile and args.web:
            url = vod_mobile + chosen_vod.getVodId()
            logger.info(f'Vod URL: {url}')
            webbrowser.open(url)
        elif args.web:
            url = vod_normal + chosen_vod.getVodId()
            logger.info(f'Vod URL: {url}')
            webbrowser.open(url)
        elif args.mobile:
            url = vod_mobile + chosen_vod.getVodId()
            command = getPlayer()
            command.append(url)
            logger.info(f'Vod Mobile: {url}')
            subprocess.Popen(command, close_fds=True)
        else:
            url = vod_normal + chosen_vod.getVodId()
            command = getPlayer()
            command.append(url)
            logger.info(f'Vod Command: {command}')
            subprocess.Popen(command, close_fds=True)

    if args.version:
        print(f'{constants.__version__}')
    elif args.vod:
        playVod()
    elif args.play or args.web:
        playWeb()
    elif args.check or args.link or args.mobile:
        checkLinkMobile()
    elif args.twitch_follows:
        streamerretriever = initStreamerRetriever()
        # print(streamerretriever.getValidate().getId())
        for f in streamerretriever.getTwitchFollows():
            print(f.getName())


def main():
    try:
        arguments()
    except KeyboardInterrupt:
        return


if __name__ == '__main__':
    main()



#! /usr/bin/env python3


import os
import re
import logging
import webbrowser
from pathlib import Path

from . import constants
from . fileoperator import FileOperator


logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)
# logger.setLevel(logging.DEBUG)

file_handler = logging.FileHandler(constants.TOOLS_LOG)
formatter = logging.Formatter(constants.TOOLS_FORM)
file_handler.setFormatter(formatter)

logger.addHandler(file_handler)


class TwitchInfo():
    """API key information"""

    __keys = {
        'name': {'twitch': 'login'},
        'user_id': {'twitch': 'user_id'},
        'expire': {'twitch': 'expires_in'}}

    def __init__(self, fo):
        self.__fo = fo
        self.__data = self.__fo.loadFile() or {}
        logger.debug('[INIT] TwitchInfo()')

    def __saveInfo(self):
        """Save data to twitch.json"""

        self.__fo.saveFile(self.__data)

    def setToken(self, token):
        """Set Twitch token"""

        self.__data['token'] = token
        logger.debug(f'[SET_TOKEN] Twitch Validation: Token {self.__data}')
        self.__saveInfo()

    def setValidate(self, data):
        """Set API key validation data"""

        self.__data['name'] = data[self.__keys['name']['twitch']]
        self.__data['user_id'] = data[self.__keys['user_id']['twitch']]
        self.__data['expire'] = data[self.__keys['expire']['twitch']]
        logger.debug(f'[SET_VALIDATE] Twitch Validation: Data {self.__data}')
        self.__saveInfo()

    def getData(self):
        """Return twitch.json"""

        logger.debug('[DATA] data return')
        return self.__data

    def getToken(self):
        """Return token from twitch.json"""

        logger.debug('[TOKEN] info return')
        return self.__data.get('token')

    def getId(self):
        """Return user id from twitch.json"""

        logger.debug('[ID] id return')
        return self.__data.get('user_id')


def oauth(funcIntpu):
    """Returns OAuth key"""

    twitch = constants.TWITCH
    base = constants.BASES[os.name]
    p = base / Path(twitch)

    fo = FileOperator(str(p))

    def __generateOAuth():
        """Open URL for Twitch OAuth key generation"""

        url = (
            'https://id.twitch.tv/oauth2/authorize'
            '?client_id=m3ggy4wqrvs4sfhv996glushskv1fi'
            '&redirect_uri=http://localhost'
            '&response_type=token')
        logger.debug(f'[OAUTH] open URL: "{url}"')
        webbrowser.open_new_tab(url)

    def getKey():
        """Return OAuth if found in file, open URL if not"""

        # oauth = fo.loadFile(singleLayer=True)
        twitch_info = TwitchInfo(fo)
        # if twitch and twitch['token']:
        #     return twitch['token']
        if twitch_info.getToken():
            logger.debug('[KEY] token found')
            return twitch_info
            # return twitch_info.getToken()
        else:
            logger.debug('[KEY] token not found')
            __generateOAuth()
            print('Paste resulting URL after doing Twitch login')
            url = funcIntpu()

            r = re.compile(
                r'^(http://localhost/)'
                r'(\#access_token=([a-z0-9]+))'
                r'(&scope=&token_type=bearer)$')
            match = r.match(url)

            if match:
                logger.debug('[KEY] token match found from URL')
                oauth = match.groups()[2]
                # fo.saveFile([oauth])
                # twitch.update({'token': oauth})
                # fo.saveFile(twitch)
                # return oauth
                twitch_info.setToken(oauth)
                return twitch_info
            else:
                logger.debug('[KEY] token not found from URL')
                raise ValueError('Given link was not recognized')

    return getKey()



#! /usr/bin/env python3


from streamerretriever import constants
import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()


setuptools.setup(
    name='Streamer Retriever',
    version=constants.__version__,
    author=constants.__author__,
    description='Filter Twitch streams that are online from follows text file.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/obtusescholar/streamerretriever',
    entry_points={ 'console_scripts': ['streamerretriever = streamerretriever.cli:main'] },
    packages=setuptools.find_packages(
        include=['streamerretriever'],
        exclude=['streamerretriever/__pycache__']
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet',
        'Topic :: Terminals',
        'Topic :: Utilities'
    ],
    license='GPLv3+',
    keywords='twitch terminal',
    python_requires='>=3.8'
)
